@@ -1,12 +1,33 @@
 /*
- * MUSCLE SmartCard Development ( http://www.linuxnet.com )
+ * MUSCLE SmartCard Development ( http://pcsclite.alioth.debian.org/pcsclite.html )
  *
  * Copyright (C) 1999-2002
- *  David Corcoran <corcoran@linuxnet.com>
+ *  David Corcoran <corcoran@musclecard.com>
  * Copyright (C) 2002-2011
  *  Ludovic Rousseau <ludovic.rousseau@free.fr>
  *
- * $Id: pcscdaemon.c 5764 2011-05-18 11:57:51Z rousseau $
+Redistribution and use in source and binary forms, with or without
+modification, are permitted provided that the following conditions
+are met:
+
+1. Redistributions of source code must retain the above copyright
+   notice, this list of conditions and the following disclaimer.
+2. Redistributions in binary form must reproduce the above copyright
+   notice, this list of conditions and the following disclaimer in the
+   documentation and/or other materials provided with the distribution.
+3. The name of the author may not be used to endorse or promote products
+   derived from this software without specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
+IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
+INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
+NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
+THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 /**
@@ -37,14 +58,15 @@
 #include "pcsclite.h"
 #include "pcscd.h"
 #include "debuglog.h"
+#include "sd-daemon.h"
 #include "winscard_msg.h"
 #include "winscard_svc.h"
 #include "sys_generic.h"
 #include "hotplug.h"
 #include "readerfactory.h"
 #include "configfile.h"
-#include "powermgt_generic.h"
 #include "utils.h"
+#include "eventhandler.h"
 
 #ifndef TRUE
 #define TRUE 1
@@ -54,19 +76,22 @@
 char AraKiri = FALSE;
 static char Init = TRUE;
 char AutoExit = FALSE;
+char SocketActivated = FALSE;
 static int ExitValue = EXIT_FAILURE;
 int HPForceReaderPolling = 0;
 static int pipefd[] = {-1, -1};
+static int signal_handler_fd[] = {-1, -1};
+char Add_Serial_In_Name = TRUE;
+char Add_Interface_In_Name = TRUE;
 
 /*
  * Some internal functions
  */
 static void at_exit(void);
 static void clean_temp_files(void);
-static void signal_reload(int sig);
 static void signal_trap(int);
-static void print_version (void);
-static void print_usage (char const * const);
+static void print_version(void);
+static void print_usage(char const * const);
 
 /**
  * @brief The Server's Message Queue Listener function.
@@ -84,6 +109,21 @@ static void SVCServiceRunLoop(void)
 
 	while (TRUE)
 	{
+		if (AraKiri)
+		{
+			/* stop the hotpug thread and waits its exit */
+#ifdef USE_USB
+			(void)HPStopHotPluggables();
+#endif
+			(void)SYS_Sleep(1);
+
+			/* now stop all the drivers */
+			RFCleanupReaders();
+			EHDeinitializeEventStructures();
+			ContextsDeinitialize();
+			at_exit();
+		}
+
 		switch (rsp = ProcessEventsServer(&dwClientID))
 		{
 
@@ -91,7 +131,7 @@ static void SVCServiceRunLoop(void)
 			Log2(PCSC_LOG_DEBUG, "A new context thread creation is requested: %d", dwClientID);
 			rv = CreateContextThread(&dwClientID);
 
- 			if (rv != SCARD_S_SUCCESS)
+			if (rv != SCARD_S_SUCCESS)
 				Log1(PCSC_LOG_ERROR, "Problem during the context thread creation");
 			break;
 
@@ -111,6 +151,10 @@ static void SVCServiceRunLoop(void)
 			/* Nothing to do in case of a syscall interrupted
 			 * It happens when SIGUSR1 (reload) or SIGINT (Ctrl-C) is received
 			 * We just try again */
+
+			/* we wait a bit so that the signal handler thread can do
+			 * its job and set AraKiri if needed */
+			SYS_USleep(1000);
 			break;
 
 		default:
@@ -118,23 +162,96 @@ static void SVCServiceRunLoop(void)
 				rsp);
 			break;
 		}
-
-		if (AraKiri)
-		{
-			/* stop the hotpug thread and waits its exit */
-#ifdef USE_USB
-			(void)HPStopHotPluggables();
-#endif
-			(void)SYS_Sleep(1);
-
-			/* now stop all the drivers */
-			RFCleanupReaders();
-			ContextsDeinitialize();
-			at_exit();
-		}
 	}
 }
 
+/**
+ * thread dedicated to handle signals
+ *
+ * a signal handler can not call any function. See signal(7) for a list
+ * of function that are safe to call from a signal handler.
+ * The functions syslog(), gettimeofday() and remove() are NOT safe.
+ */
+static void *signal_thread(void *arg)
+{
+	(void)arg;
+
+	while (TRUE)
+	{
+		int r;
+		int sig;
+
+		r = read(signal_handler_fd[0], &sig, sizeof sig);
+		if (r < 0)
+		{
+			Log2(PCSC_LOG_ERROR, "read failed: %s", strerror(errno));
+			return NULL;
+		}
+
+		Log2(PCSC_LOG_INFO, "Received signal: %d", sig);
+
+		/* signal for hotplug */
+		if (SIGUSR1 == sig)
+		{
+#ifdef USE_USB
+			if (! AraKiri)
+				HPReCheckSerialReaders();
+#endif
+			/* Reenable the signal handler.
+			 * This is needed on Solaris and HPUX. */
+			(void)signal(SIGUSR1, signal_trap);
+
+			continue;
+		}
+
+		/* do not wait if asked to terminate
+		 * avoids waiting after the reader(s) in shutdown for example */
+		if (SIGTERM == sig)
+		{
+			Log1(PCSC_LOG_INFO, "Direct suicide");
+			at_exit();
+		}
+
+		if (SIGALRM == sig)
+		{
+			/* normal exit without error */
+			ExitValue = EXIT_SUCCESS;
+		}
+
+		/* the signal handler is called several times for the same Ctrl-C */
+		if (AraKiri == FALSE)
+		{
+			Log1(PCSC_LOG_INFO, "Preparing for suicide");
+			AraKiri = TRUE;
+
+			/* if still in the init/loading phase the AraKiri will not be
+			 * seen by the main event loop
+			 */
+			if (Init)
+			{
+				Log1(PCSC_LOG_INFO, "Suicide during init");
+				at_exit();
+			}
+		}
+		else
+		{
+			/* if pcscd do not want to die */
+			static int lives = 2;
+
+			lives--;
+			/* no live left. Something is blocking the normal death. */
+			if (0 == lives)
+			{
+				Log1(PCSC_LOG_INFO, "Forced suicide");
+				at_exit();
+			}
+		}
+	}
+
+	return NULL;
+}
+
+
 int main(int argc, char **argv)
 {
 	int rv;
@@ -146,7 +263,7 @@ int main(int argc, char **argv)
 	int customMaxReaderHandles = 0;
 	int customMaxThreadCardHandles = 0;
 	int opt;
-	int limited_rights = FALSE;
+	int r;
 #ifdef HAVE_GETOPT_LONG
 	int option_index = 0;
 	static struct option long_options[] = {
@@ -157,7 +274,7 @@ int main(int argc, char **argv)
 		{"version", 0, NULL, 'v'},
 		{"apdu", 0, NULL, 'a'},
 		{"debug", 0, NULL, 'd'},
-		{"info", 0, NULL, 0},
+		{"info", 0, NULL, 'i'},
 		{"error", 0, NULL, 'e'},
 		{"critical", 0, NULL, 'C'},
 		{"hotplug", 0, NULL, 'H'},
@@ -166,10 +283,12 @@ int main(int argc, char **argv)
 		{"max-card-handle-per-thread", 1, NULL, 's'},
 		{"max-card-handle-per-reader", 1, NULL, 'r'},
 		{"auto-exit", 0, NULL, 'x'},
+		{"reader-name-no-serial", 0, NULL, 'S'},
+		{"reader-name-no-interface", 0, NULL, 'I'},
 		{NULL, 0, NULL, 0}
 	};
 #endif
-#define OPT_STRING "c:fTdhvaeCHt:r:s:x"
+#define OPT_STRING "c:fTdhvaieCHt:r:s:xSI"
 
 	newReaderConfig = NULL;
 	setToForeground = FALSE;
@@ -188,15 +307,15 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 
+	/* Init the PRNG */
+	SYS_InitRandom();
+
 	/*
 	 * By default we create a daemon (not connected to any output)
 	 * so log to syslog to have error messages.
 	 */
 	DebugLogSetLogType(DEBUGLOG_SYSLOG_DEBUG);
 
-	/* if the process is setuid or setgid it may have some restrictions */
-	limited_rights = (getgid() != getegid()) && (getuid() != 0);
-
 	/*
 	 * Handle any command line arguments
 	 */
@@ -214,11 +333,6 @@ int main(int argc, char **argv)
 				break;
 #endif
 			case 'c':
-				if (limited_rights)
-				{
-					Log1(PCSC_LOG_CRITICAL, "Can't use a user specified config file");
-					return EXIT_FAILURE;
-				}
 				Log2(PCSC_LOG_INFO, "using new config file: %s", optarg);
 				newReaderConfig = optarg;
 				break;
@@ -240,6 +354,10 @@ int main(int argc, char **argv)
 				DebugLogSetLevel(PCSC_LOG_DEBUG);
 				break;
 
+			case 'i':
+				DebugLogSetLevel(PCSC_LOG_INFO);
+				break;
+
 			case 'e':
 				DebugLogSetLevel(PCSC_LOG_ERROR);
 				break;
@@ -257,12 +375,7 @@ int main(int argc, char **argv)
 				return EXIT_SUCCESS;
 
 			case 'a':
-				if (limited_rights)
-				{
-					Log1(PCSC_LOG_CRITICAL, "Can't log APDU (restricted)");
-					return EXIT_FAILURE;
-				}
-				(void)DebugLogSetCategory(DEBUG_CATEGORY_APDU);
+				DebugLogSetCategory(DEBUG_CATEGORY_APDU);
 				break;
 
 			case 'H':
@@ -273,24 +386,18 @@ int main(int argc, char **argv)
 
 			case 't':
 				customMaxThreadCounter = optarg ? atoi(optarg) : 0;
-				if (limited_rights && (customMaxThreadCounter < PCSC_MAX_CONTEXT_THREADS))
-					customMaxThreadCounter = PCSC_MAX_CONTEXT_THREADS;
 				Log2(PCSC_LOG_INFO, "setting customMaxThreadCounter to: %d",
 					customMaxThreadCounter);
 				break;
 
 			case 'r':
 				customMaxReaderHandles = optarg ? atoi(optarg) : 0;
-				if (limited_rights && (customMaxReaderHandles < PCSC_MAX_READER_HANDLES))
-					customMaxReaderHandles = PCSC_MAX_READER_HANDLES;
 				Log2(PCSC_LOG_INFO, "setting customMaxReaderHandles to: %d",
 					customMaxReaderHandles);
 				break;
 
 			case 's':
 				customMaxThreadCardHandles = optarg ? atoi(optarg) : 0;
-				if (limited_rights && (customMaxThreadCardHandles < PCSC_MAX_CONTEXT_CARD_HANDLES))
-					customMaxThreadCardHandles = PCSC_MAX_CONTEXT_CARD_HANDLES;
 				Log2(PCSC_LOG_INFO, "setting customMaxThreadCardHandles to: %d",
 					customMaxThreadCardHandles);
 				break;
@@ -301,6 +408,14 @@ int main(int argc, char **argv)
 					TIME_BEFORE_SUICIDE);
 				break;
 
+			case 'S':
+				Add_Serial_In_Name = FALSE;
+				break;
+
+			case 'I':
+				Add_Interface_In_Name = FALSE;
+				break;
+
 			default:
 				print_usage (argv[0]);
 				return EXIT_FAILURE;
@@ -315,13 +430,34 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 
+	/*
+	 * Check if systemd passed us any file descriptors
+	 */
+	rv = sd_listen_fds(0);
+	if (rv > 1)
+	{
+		Log1(PCSC_LOG_CRITICAL, "Too many file descriptors received");
+		return EXIT_FAILURE;
+	}
+	else
+	{
+		if (rv == 1)
+		{
+			SocketActivated = TRUE;
+			Log1(PCSC_LOG_INFO, "Started by systemd");
+		}
+		else
+			SocketActivated = FALSE;
+	}
+
 	/*
 	 * test the presence of /var/run/pcscd/pcscd.comm
 	 */
 
 	rv = stat(PCSCLITE_CSOCK_NAME, &fStatBuf);
 
-	if (rv == 0)
+	/* if the file exist and pcscd was _not_ started by systemd */
+	if (rv == 0 && !SocketActivated)
 	{
 		pid_t pid;
 
@@ -341,7 +477,7 @@ int main(int argc, char **argv)
 				Log1(PCSC_LOG_CRITICAL,
 					"file " PCSCLITE_CSOCK_NAME " already exists.");
 				Log2(PCSC_LOG_CRITICAL,
-					"Another pcscd (pid: %d) seems to be running.", pid);
+					"Another pcscd (pid: %ld) seems to be running.", (long)pid);
 				return EXIT_FAILURE;
 			}
 			else
@@ -365,17 +501,6 @@ int main(int argc, char **argv)
 				Log1(PCSC_LOG_CRITICAL, "Hotplug failed");
 				return EXIT_FAILURE;
 			}
-
-			Log1(PCSC_LOG_CRITICAL,
-				"file " PCSCLITE_CSOCK_NAME " already exists.");
-			Log1(PCSC_LOG_CRITICAL,
-				"Maybe another pcscd is running?");
-			Log1(PCSC_LOG_CRITICAL,
-				"I can't read process pid from " PCSCLITE_RUN_PID);
-			Log1(PCSC_LOG_CRITICAL, "Remove " PCSCLITE_CSOCK_NAME);
-			Log1(PCSC_LOG_CRITICAL,
-				"if pcscd is not running to clear this message.");
-			return EXIT_FAILURE;
 		}
 	}
 	else
@@ -387,24 +512,11 @@ int main(int argc, char **argv)
 
 	/* like in daemon(3): changes the current working directory to the
 	 * root ("/") */
-	(void)chdir("/");
-
-	if (AutoExit)
+	r = chdir("/");
+	if (r < 0)
 	{
-		int pid;
-
-		/* create a new session so that Ctrl-C on the application will
-		 * not also quit pcscd */
-		setsid();
-
-		/* fork() so that pcscd always return in --auto-exit mode */
-		pid = fork();
-		if (-1 == pid )
-			Log2(PCSC_LOG_CRITICAL, "fork() failed: %s", strerror(errno));
-
-		if (pid)
-			/* father */
-			return EXIT_SUCCESS;
+		Log2(PCSC_LOG_CRITICAL, "chdir() failed: %s", strerror(errno));
+		return EXIT_FAILURE;
 	}
 
 	/*
@@ -413,6 +525,7 @@ int main(int argc, char **argv)
 	if (!setToForeground)
 	{
 		int pid;
+		int fd;
 
 		if (pipe(pipefd) == -1)
 		{
@@ -429,9 +542,17 @@ int main(int argc, char **argv)
 
 		/* like in daemon(3): redirect standard input, standard output
 		 * and standard error to /dev/null */
-		(void)close(0);
-		(void)close(1);
-		(void)close(2);
+		fd = open("/dev/null", O_RDWR);
+		if (fd != -1)
+		{
+			dup2(fd, STDIN_FILENO);
+			dup2(fd, STDOUT_FILENO);
+			dup2(fd, STDERR_FILENO);
+
+			/* do not close stdin, stdout or stderr */
+			if (fd > 2)
+				close(fd);
+		}
 
 		if (pid)
 		/* in the father */
@@ -471,16 +592,28 @@ int main(int argc, char **argv)
 	/* exits on SIGALARM to allow pcscd to suicide if not used */
 	(void)signal(SIGALRM, signal_trap);
 
+	if (pipe(signal_handler_fd) == -1)
+	{
+		Log2(PCSC_LOG_CRITICAL, "pipe() failed: %s", strerror(errno));
+		return EXIT_FAILURE;
+	}
+
+	pthread_t signal_handler_thread;
+	rv = pthread_create(&signal_handler_thread, NULL, signal_thread, NULL);
+	if (rv)
+	{
+		Log2(PCSC_LOG_CRITICAL, "pthread_create failed: %s", strerror(rv));
+		return EXIT_FAILURE;
+	}
+
 	/*
 	 * If PCSCLITE_IPC_DIR does not exist then create it
 	 */
-	rv = stat(PCSCLITE_IPC_DIR, &fStatBuf);
-	if (rv < 0)
 	{
 		int mode = S_IROTH | S_IXOTH | S_IRGRP | S_IXGRP | S_IRWXU;
 
 		rv = mkdir(PCSCLITE_IPC_DIR, mode);
-		if (rv != 0)
+		if ((rv != 0) && (errno != EEXIST))
 		{
 			Log2(PCSC_LOG_CRITICAL,
 				"cannot create " PCSCLITE_IPC_DIR ": %s", strerror(errno));
@@ -493,37 +626,6 @@ int main(int argc, char **argv)
 		(void)chmod(PCSCLITE_IPC_DIR, mode);
 	}
 
-	/*
-	 * Record our pid to make it easier
-	 * to kill the correct pcscd
-	 */
-	{
-		int f;
-		int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
-
-		f = open(PCSCLITE_RUN_PID, O_RDWR | O_CREAT, mode);
-		if (f != -1)
-		{
-			char pid[PID_ASCII_SIZE];
-
-			(void)snprintf(pid, sizeof(pid), "%u\n", (unsigned) getpid());
-			(void)write(f, pid, strlen(pid));
-			(void)close(f);
-
-			/* set mode so that the file is world readable even is umask is
-			 * restrictive
-			 * The file is used by libpcsclite */
-			(void)chmod(PCSCLITE_RUN_PID, mode);
-		}
-		else
-			Log2(PCSC_LOG_CRITICAL, "cannot create " PCSCLITE_RUN_PID ": %s",
-				strerror(errno));
-	}
-
-	/* cleanly remove /var/run/pcscd/pcsc.* files when exiting */
-	if (atexit(at_exit))
-		Log2(PCSC_LOG_CRITICAL, "atexit() failed: %s", strerror(errno));
-
 	/*
 	 * Allocate memory for reader structures
 	 */
@@ -555,6 +657,42 @@ int main(int argc, char **argv)
 
 	Log1(PCSC_LOG_INFO, "pcsc-lite " VERSION " daemon ready.");
 
+	/*
+	 * Record our pid to make it easier
+	 * to kill the correct pcscd
+	 *
+	 * Do not fork after this point or the stored pid will be wrong
+	 */
+	{
+		int f;
+		int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
+
+		f = open(PCSCLITE_RUN_PID, O_RDWR | O_CREAT, mode);
+		if (f != -1)
+		{
+			char pid[PID_ASCII_SIZE];
+			ssize_t rr;
+
+			(void)snprintf(pid, sizeof(pid), "%u\n", (unsigned) getpid());
+			rr = write(f, pid, strlen(pid) + 1);
+			if (rr < 0)
+			{
+				Log2(PCSC_LOG_CRITICAL,
+					"writing " PCSCLITE_RUN_PID " failed: %s",
+					strerror(errno));
+			}
+			(void)close(f);
+
+			/* set mode so that the file is world readable even is umask is
+			 * restrictive
+			 * The file is used by libpcsclite */
+			(void)chmod(PCSCLITE_RUN_PID, mode);
+		}
+		else
+			Log2(PCSC_LOG_CRITICAL, "cannot create " PCSCLITE_RUN_PID ": %s",
+				strerror(errno));
+	}
+
 	/*
 	 * post initialistion
 	 */
@@ -563,12 +701,16 @@ int main(int argc, char **argv)
 	/*
 	 * Hotplug rescan
 	 */
-	(void)signal(SIGUSR1, signal_reload);
+	(void)signal(SIGUSR1, signal_trap);
 
 	/*
 	 * Initialize the comm structure
 	 */
-	rv = InitializeSocket();
+	if (SocketActivated)
+		rv = ListenExistingSocket(SD_LISTEN_FDS_START + 0);
+	else
+		rv = InitializeSocket();
+
 	if (rv)
 	{
 		Log1(PCSC_LOG_CRITICAL, "Error initializing pcscd.");
@@ -595,8 +737,10 @@ int main(int argc, char **argv)
 	 * Set up the search for USB/PCMCIA devices
 	 */
 	rv = HPSearchHotPluggables();
+#ifndef USE_SERIAL
 	if (rv)
 		at_exit();
+#endif
 
 	rv = HPRegisterForHotplugEvents();
 	if (rv)
@@ -608,19 +752,20 @@ int main(int argc, char **argv)
 	RFWaitForReaderInit();
 #endif
 
-	/*
-	 * Set up the power management callback routine
-	 */
-	(void)PMRegisterForPowerEvents();
-
 	/* initialisation succeeded */
 	if (pipefd[1] >= 0)
 	{
 		char buf = 0;
+		ssize_t rr;
 
 		/* write a 0 (success) to father process */
-		write(pipefd[1], &buf, 1);
+		rr = write(pipefd[1], &buf, 1);
+		if (rr < 0)
+		{
+			Log2(PCSC_LOG_ERROR, "write() failed: %s", strerror(errno));
+		}
 		close(pipefd[1]);
+		pipefd[1] = -1;
 	}
 
 	SVCServiceRunLoop();
@@ -638,10 +783,15 @@ static void at_exit(void)
 	if (pipefd[1] >= 0)
 	{
 		char buf;
+		ssize_t r;
 
 		/* write the error code to father process */
 		buf = ExitValue;
-		write(pipefd[1], &buf, 1);
+		r = write(pipefd[1], &buf, 1);
+		if (r < 0)
+		{
+			Log2(PCSC_LOG_ERROR, "write() failed: %s", strerror(errno));
+		}
 		close(pipefd[1]);
 	}
 
@@ -652,10 +802,13 @@ static void clean_temp_files(void)
 {
 	int rv;
 
-	rv = remove(PCSCLITE_CSOCK_NAME);
-	if (rv != 0)
-		Log2(PCSC_LOG_ERROR, "Cannot remove " PCSCLITE_CSOCK_NAME ": %s",
-			strerror(errno));
+	if (!SocketActivated)
+	{
+		rv = remove(PCSCLITE_CSOCK_NAME);
+		if (rv != 0)
+			Log2(PCSC_LOG_ERROR, "Cannot remove " PCSCLITE_CSOCK_NAME ": %s",
+				strerror(errno));
+	}
 
 	rv = remove(PCSCLITE_RUN_PID);
 	if (rv != 0)
@@ -663,74 +816,27 @@ static void clean_temp_files(void)
 			strerror(errno));
 }
 
-static void signal_reload(/*@unused@*/ int sig)
-{
-	(void)signal(SIGUSR1, signal_reload);
-
-	(void)sig;
-
-	if (AraKiri)
-		return;
-
-#ifdef USE_USB
-	HPReCheckSerialReaders();
-#endif
-} /* signal_reload */
-
 static void signal_trap(int sig)
 {
-	Log2(PCSC_LOG_INFO, "Received signal: %d", sig);
+	int r;
 
-	/* do not wait if asked to terminate
-	 * avoids waiting after the reader(s) in shutdown for example */
-	if (SIGTERM == sig)
-	{
-		Log1(PCSC_LOG_INFO, "Direct suicide");
-		at_exit();
-	}
-
-	/* the signal handler is called several times for the same Ctrl-C */
-	if (AraKiri == FALSE)
-	{
-		Log1(PCSC_LOG_INFO, "Preparing for suicide");
-		AraKiri = TRUE;
-
-		/* if still in the init/loading phase the AraKiri will not be
-		 * seen by the main event loop
-		 */
-		if (Init)
-		{
-			Log1(PCSC_LOG_INFO, "Suicide during init");
-			at_exit();
-		}
-	}
-	else
-	{
-		/* if pcscd do not want to die */
-		static int lives = 2;
-
-		lives--;
-		/* no live left. Something is blocking the normal death. */
-		if (0 == lives)
-		{
-			Log1(PCSC_LOG_INFO, "Forced suicide");
-			at_exit();
-		}
-	}
+	r = write(signal_handler_fd[1], &sig, sizeof sig);
+	if (r < 0)
+		Log2(PCSC_LOG_ERROR, "write failed: %s", strerror(errno));
 }
 
-static void print_version (void)
+static void print_version(void)
 {
 	printf("%s version %s.\n",  PACKAGE, VERSION);
-	printf("Copyright (C) 1999-2002 by David Corcoran <corcoran@linuxnet.com>.\n");
-	printf("Copyright (C) 2001-2011 by Ludovic Rousseau <ludovic.rousseau@free.fr>.\n");
+	printf("Copyright (C) 1999-2002 by David Corcoran <corcoran@musclecard.com>.\n");
+	printf("Copyright (C) 2001-2015 by Ludovic Rousseau <ludovic.rousseau@free.fr>.\n");
 	printf("Copyright (C) 2003-2004 by Damien Sauveron <sauveron@labri.fr>.\n");
-	printf("Report bugs to <muscle@lists.musclecard.com>.\n");
+	printf("Report bugs to <pcsclite-muscle@lists.alioth.debian.org>.\n");
 
 	printf ("Enabled features:%s\n", PCSCLITE_FEATURES);
 }
 
-static void print_usage (char const * const progname)
+static void print_usage(char const * const progname)
 {
 	printf("Usage: %s options\n", progname);
 	printf("Options:\n");
@@ -739,28 +845,33 @@ static void print_usage (char const * const progname)
 	printf("  -c, --config		path to reader.conf\n");
 	printf("  -f, --foreground	run in foreground (no daemon),\n");
 	printf("			send logs to stdout instead of syslog\n");
-	printf("  -T, --color       force use of colored logs\n");
+	printf("  -T, --color		force use of colored logs\n");
 	printf("  -h, --help		display usage information\n");
 	printf("  -H, --hotplug		ask the daemon to rescan the available readers\n");
 	printf("  -v, --version		display the program version number\n");
-	printf("  -d, --debug	 	display lower level debug messages\n");
-	printf("      --info	 	display info level debug messages (default level)\n");
-	printf("  -e  --error	 	display error level debug messages\n");
-	printf("  -C  --critical 	display critical only level debug messages\n");
+	printf("  -d, --debug		display lower level debug messages\n");
+	printf("  -i, --info		display info level debug messages\n");
+	printf("  -e  --error		display error level debug messages (default level)\n");
+	printf("  -C  --critical	display critical only level debug messages\n");
 	printf("  --force-reader-polling ignore the IFD_GENERATE_HOTPLUG reader capability\n");
 	printf("  -t, --max-thread	maximum number of threads (default %d)\n", PCSC_MAX_CONTEXT_THREADS);
 	printf("  -s, --max-card-handle-per-thread	maximum number of card handle per thread (default: %d)\n", PCSC_MAX_CONTEXT_CARD_HANDLES);
 	printf("  -r, --max-card-handle-per-reader	maximum number of card handle per reader (default: %d)\n", PCSC_MAX_READER_HANDLES);
-	printf("  -x, --auto-exits	pcscd will quit after %d seconds of inactivity\n", TIME_BEFORE_SUICIDE);
+	printf("  -x, --auto-exit	pcscd will quit after %d seconds of inactivity\n", TIME_BEFORE_SUICIDE);
+	printf("  -S, --reader-name-no-serial    do not include the USB serial number in the name\n");
+	printf("  -I, --reader-name-no-interface do not include the USB interface name in the name\n");
 #else
 	printf("  -a    log APDU commands and results\n");
-	printf("  -c 	path to reader.conf\n");
+	printf("  -c	path to reader.conf\n");
 	printf("  -f	run in foreground (no daemon), send logs to stdout instead of syslog\n");
 	printf("  -T    force use of colored logs\n");
-	printf("  -d 	display debug messages. Output may be:\n");
-	printf("  -h 	display usage information\n");
+	printf("  -d	display debug messages.\n");
+	printf("  -i	display info messages.\n");
+	printf("  -e	display error messages (default level).\n");
+	printf("  -C	display critical messages.\n");
+	printf("  -h	display usage information\n");
 	printf("  -H	ask the daemon to rescan the available readers\n");
-	printf("  -v 	display the program version number\n");
+	printf("  -v	display the program version number\n");
 	printf("  -t    maximum number of threads\n");
 	printf("  -s    maximum number of card handle per thread\n");
 	printf("  -r    maximum number of card handle per reader\n");
